@@ -2,15 +2,17 @@ from curses import tigetstr, setupterm, tparm
 from fcntl import ioctl
 from functools import partial
 from itertools import cycle
-import os
+from os import getcwd
+from os.path import abspath
 from signal import signal, SIGWINCH
 import struct
 from termios import TIOCGWINSZ
+from time import time
 from traceback import format_list, extract_tb, format_exception_only
 
 from nose.plugins import Plugin
 from nose.exc import SkipTest
-from nose.util import test_address
+from nose.util import test_address, src
 
 
 class ProgressivePlugin(Plugin):
@@ -68,6 +70,15 @@ class ProgressivePlugin(Plugin):
                                                 loader.loadTestsFromNames)
         return loader
 
+    def prepareTestRunner(self, runner):
+        """Store start time so finalize() can use it.
+
+        This is as close to the start of the run as we can get without being
+        our own testrunner.
+
+        """
+        self._startTime = time()
+
     def printError(self, kind, err, test):
         """Output a human-readable error report to the stream.
 
@@ -84,12 +95,11 @@ class ProgressivePlugin(Plugin):
         # TODO: Canonicalize the path to remove /kitsune/../kitsune nonsense.
         # Don't relativize, though, as that hurts the ability to paste into
         # running editors.
+        writeln = self.stream.writeln
+        write = self.stream.write
         with self.bar.dodging():
-            writeln = self.stream.writeln
-            write = self.stream.write
-            writeln()
-            write(tigetstr('bold'))
-            writeln('%s: %s' % (kind, nose_selector(test)))
+            writeln('\n' + tigetstr('bold') +
+                  '%s: %s' % (kind, nose_selector(test)))
 
             # File name and line num in a format vi can take:
             address = test.address()
@@ -97,21 +107,16 @@ class ProgressivePlugin(Plugin):
                 file_name = address[0]
                 if file_name:
                     line_num = extracted_tb[-1][1]
-                    writeln(' ' * len(kind) +
-                            '  %s +%s' % (relative_source_path(file_name), line_num))
+                    writeln(' ' * len(kind) + '  +%s %s' %
+                            (line_num, human_path(src(file_name))))
 
-            write(tigetstr('sgr0'))
+            write(tigetstr('sgr0'))  # end bold
 
             # Traceback:
-            self.stream.write(formatted_err)
+            write(formatted_err)
 
             # Exception:
-            self.stream.write(''.join(format_exception_only(exception_type, exception_value)))
-
-    def printErrors(self):
-        # The current summary doesn't begin with a \n.
-        with self.bar.dodging():
-            self.stream.writeln()
+            write(''.join(format_exception_only(exception_type, exception_value)))
 
     def finalize(self, result):
         """Print counts of tests run."""
@@ -121,12 +126,16 @@ class ProgressivePlugin(Plugin):
             types.append('skip')
             values.append(len(result.skipped))
         msg = ', '.join('%s %s%s' % (v, t, 's' if v != 1 else '')
-                        for t, v in zip(types, values))
+                        for t, v in zip(types, values)) + \
+              ' in %.1fs' % (time() - self._startTime)
 
         # Erase progress bar. Bash doesn't clear the whole line when printing
         # the prompt, leaving a piece of the bar.
         self.bar.erase()
-        self.stream.writeln('\n' + msg)
+        self.stream.writeln()
+        if not result.failures and not result.errors:
+            self.stream.write('OK!  ')
+        self.stream.writeln(msg)
 
     def startTest(self, test):
         # Overriding this seems to prevent TextTestRunner from running its, so
@@ -259,12 +268,15 @@ def nose_selector(test):
         return module
 
 
-def relative_source_path(path):
-    """Return a relative filesystem path to the source file of the possibly-compiled module at the given path."""
-    path = os.path.abspath(path)
-    cwd = os.getcwd()
+def human_path(path):
+    """Return the most human-readable representation of the given path.
+    
+    If an absolute path is given that's within the current directory, convert
+    it to a relative path to shorten it. Otherwise, return the absolute path.
+    
+    """
+    path = abspath(path)
+    cwd = getcwd()
     if path.startswith(cwd):
         path = path[len(cwd) + 1:]  # Make path relative. Remove leading slash.
-    if path.endswith('.pyc'):
-        path = path[:-1]
     return path
